// LogEntry.cpp -- Implementation file for the LogEntry class
//                 The LogEntry class describes an http log file entry
// Written by Dr. Chris Taylor, t a y l o r@m s o e.edu
// Date: 2-15-2003
// Note: I have included an excessive amount of comments here to help
//  clarify what is going on so that students unfamiliar with some of
//  these techniques can learn from them.
// Modified: 3-11-2003, t a y l o r@m s o e.edu: added getStatus()

#include <fstream>
#include <string>
#include <sstream>
#include <iomanip>
#include "LogEntry.h"

using std::string;
using std::ostream;
using std::istream;
using std::setfill;
using std::setw;
using std::istringstream;
using std::getline;

//Define a static const variable to the largest possible value
//  for an unsigned int.  This value is used to signify invalid data.
const unsigned int LogEntry::invalid = static_cast<unsigned int>(-1);

// t a y l o r@m s o e.edu, 2-15-2003
LogEntry::LogEntry() : host("-"), identity("-"), userid("-"),
                       day(invalid), month("-"), year(invalid),
                       hour(invalid), minute(invalid), second(invalid),
                       offset("-"), clientRequest("-"),
                       status(invalid), size(invalid),
                       referrer("-"), userAgent("-")
{
  // Nothing else to do
}

// t a y l o r@m s o e.edu, 2-15-2003
LogEntry::LogEntry(const LogEntry& org) : host(org.host), identity(org.identity),
                                          userid(org.userid), day(org.day),
                                          month(org.month), year(org.year),
                                          hour(org.hour), minute(org.minute),
                                          second(org.second), offset(org.offset),
                                          clientRequest(org.clientRequest),
                                          status(org.status), size(org.size),
                                          userAgent(org.userAgent)
{
  // Nothing else to do
}

// t a y l o r@m s o e.edu, 2-15-2003
LogEntry::~LogEntry()
{
  // Nothing to do
}

// t a y l o r@m s o e.edu, 2-15-2003
LogEntry& LogEntry::operator=(const LogEntry& rhs)
{
  // Check for self-assignment
  if(&rhs!=this) {
    host = rhs.host;
    identity = rhs.identity;
    userid = rhs.userid;
    day = rhs.day;
    month = rhs.month;
    year = rhs.year,
    hour = rhs.hour;
    minute = rhs.minute,
    second = rhs.second;
    offset = rhs.offset;
    clientRequest = rhs.clientRequest;
    status = rhs.status;
    size = rhs.size;
    referrer = rhs.referrer;
    userAgent = rhs.userAgent;
  }
  return *this;
}

// t a y l o r@m s o e.edu, 2-15-2003
bool LogEntry::extract(const string& entry)
{
  bool successful = true;
  // save original values in case there is an error on extraction
  LogEntry temp(*this);
  {
    LogEntry empty;
    *this = empty;
  }
  // Convert the string to an input string stream and use the extraction
  //  operator to parse the string
  istringstream iss(entry);
  char junk;
  // iss will evaluate to true as long as no error flags have been set.
  //  If an error occurs trying to extract the host, identity, userid,
  //  or day, an error flag will be set and our successful flag will
  //  indicate an error in parsing the input stream
  //  Similarly, if the character preceeding the day (junk) is not a
  //  left square bracket ([), the successful flag will indicate an
  //  error since the input stream was not what we expected
  if(!(iss>>host>>identity>>userid>>junk>>day) || '['!=junk) {
    successful = false;
  } else {
    string dateTime;
    // Here we use a version of getline which will read everything in
    //  the input stream up to a right square bracket (]).  This is
    //  an easy way to get all of the date information into a string
    //  to be parsed separately.
    // In addition, we are verifying that each of the separating characters
    //  is what we expect.  Before we checked these, though, we must make
    //  sure the string is sufficently long.  If we didn't do this, and
    //  the string received from getline was only 3 characters long, we
    //  would be indexing off the end of the legal string, and this could
    //  cause our program to crash.
    if(getline(iss, dateTime, ']') && (dateTime.size()>17) &&
       ('/'==dateTime[0]) && ('/'==dateTime[4]) && (':'==dateTime[9]) &&
       (':'==dateTime[12]) && (':'==dateTime[15])) {
      month = dateTime.substr(1,3);
      // Each of these extractions is done in its own code block to
      //  limit the lifetime of the input string stream object, iss.
      //  Be sure to note that these are different iss objects than
      //  the one used above.  The iss objects below are specific to
      //  the code block in which they are declared.
      {
        istringstream iss(dateTime.substr(5,4));
        if(!(iss>>year)) {
          successful = false;
        }
      }
      {
        istringstream iss(dateTime.substr(10,2));
        if(!(iss>>hour)) {
          successful = false;
        }
      }
      {
        istringstream iss(dateTime.substr(13,2));
        if(!(iss>>minute)) {
          successful = false;
        }
      }
      {
        istringstream iss(dateTime.substr(16,2));
        if(!(iss>>second)) {
          successful = false;
        }
      }
      // The offset is an optional parameter so we want to be sure to
      //  handle the situation where it is present and also where it is not.
      if(dateTime.size()>19) {
        offset=dateTime.substr(19,5);
      } else {
        offset="";
      }
    } else {
      successful = false;
    }
  }
  if(!successful || !(iss>>junk) || '\"'!=junk || !(getline(iss, clientRequest, '\"')) ||
     !(iss>>status)) {
    successful = false;
  } else {
    // The size could be a character (if the file requested doesn't exist)
    //  so we've got some extra hoops to jump through... if the requested
    //  file doesn't exist, we'll set size to LogEntry::invalid
    string tempsize;
    if(iss>>tempsize) {
      if(tempsize.size()!=0 && tempsize[0]=='-') {
        size = invalid;
      } else {
        // Note: this is a different iss that is local to this else statement
        istringstream iss(tempsize);
        if(!(iss>>size)) {
          successful = false;
        }
      }
    } else {
      successful = false;
    }
    // If we made it this far, we have been successful.  We'll try to get the
    //  last two fields, but if they don't exist, that's okay because they are
    //  optional.
    iss>>junk;
    getline(iss, referrer, '\"');
    iss>>junk;
    getline(iss, userAgent, '\"');
    if(""==referrer) {
      referrer="-";
    }
    if(""==userAgent) {
      userAgent="-";
    }
  }
  // If we ran into trouble, we restore the original value of the object
  if(!successful) {
    *this = temp;
  }
  return successful;
}

// t a y l o r@m s o e.edu, 2-15-2003
void LogEntry::insert(ostream& os) const
{
  // We use setfill and setw to make sure that the day, hour, minute,
  //  and second that we place in the output stream is two digits long.
  //  (05:17:02) instead of (5:17:2)
  os << host << ' '
     << identity << ' '
     << userid << " [" << setfill('0')
     << setw(2) << day << '/'
     << month << '/'
     << year << ':'
     << setw(2) << hour << ':'
     << setw(2) << minute << ':'
     << setw(2) << second;
  // Insert the offset into the stream only if it exists
  if(offset.size()!=0) {
    os << ' ' << offset;
  }
  os << "] \""
     << clientRequest << "\" "
     << status << ' ';
  // Insert '-' for size if the size is invalid
  if(invalid==size) {
    os << '-';
  } else {
    os << size;
  }
  // Insert referrer and userAgent only if they exist
  if("-"!=referrer) {
    os << " \"" << referrer << "\"";
    if("-"!=userAgent) {
      os << " \"" << userAgent << "\"";
    }
  }
}

// t a y l o r@m s o e.edu, 2-15-2003
string LogEntry::getHost() const
{
  return host;
}

// t a y l o r@m s o e.edu, 2-15-2003
string LogEntry::getIdentity() const
{
  return identity;
}

// t a y l o r@m s o e.edu, 2-15-2003
string LogEntry::getUserid() const
{
  return userid;
}

// t a y l o r@m s o e.edu, 2-15-2003
unsigned int LogEntry::getDay() const
{
  return day;
}

// t a y l o r@m s o e.edu, 2-15-2003
string LogEntry::getMonth() const
{
  return month;
}

// t a y l o r@m s o e.edu, 2-15-2003
unsigned int LogEntry::getYear() const
{
  return year;
}

// t a y l o r@m s o e.edu, 2-15-2003
unsigned int LogEntry::getHour() const
{
  return hour;
}

// t a y l o r@m s o e.edu, 2-15-2003
unsigned int LogEntry::getMinute() const
{
  return minute;
}

// t a y l o r@m s o e.edu, 2-15-2003
unsigned int LogEntry::getSecond() const
{
  return second;
}

// t a y l o r@m s o e.edu, 2-15-2003
unsigned int LogEntry::getSize() const
{
  return size;
}

// t a y l o r@m s o e.edu, 2-15-2003
string LogEntry::getReferrer() const
{
  return referrer;
}

// t a y l o r@m s o e.edu, 2-15-2003
string LogEntry::getUserAgent() const
{
  return userAgent;
}

// t a y l o r@m s o e.edu, 3-11-2003
unsigned int LogEntry::getStatus() const
{
  return status;
}

// t a y l o r@m s o e.edu, 2-15-2003
ostream& operator<<(ostream& os, const LogEntry& entry)
{
  entry.insert(os);
  return os;
}

// t a y l o r@m s o e.edu, 2-15-2003
istream& operator>>(istream& is, LogEntry& entry)
{
  bool successful=false;
  string line;
  if(getline(is, line)) {
    successful = entry.extract(line);
  }
  // If the LogEntry::extract function fails, we need to set the ios::failbit
  if(!successful) {
    is.setstate(std::ios::failbit);
  }
  return is;
}
